#ifndef KCS_H
#define KCS_H
#include "icc_what.h"
#include "kcs_ifc.h"


/*VSC452 KCS Registers */
#define SYS_CNTL_BASE 			0xB0200000
#define KCS0_BASE     			(SYS_CNTL_BASE + 0x0D00)
#define KCS1_BASE			(SYS_CNTL_BASE + 0x0D80)
#define KCS2_BASE     			(SYS_CNTL_BASE + 0x0E00)
#define KCS3_BASE			(SYS_CNTL_BASE + 0x0E80)



#define IRQ_KCS_BT_SMIC_MAIL	2	


/* as of now even 512 is too much for us*/
#define MSG_PAYLOAD_SIZE	(512)

/* status confliction with local variables, so better change it to STATUS_REG*/
#define STATUS_REG		(0x01) 
#define DATA_IN   	(0x00)
#define CMD_IN		(0x02)
#define DATA_OUT  	(0x00)

/* VSC452 Specific Mode register */
#define MODE		(0x03)



#define	KCS_1_CHANNEL			1
#define	KCS_2_CHANNEL			2
#define	KCS_3_CHANNEL			3
#define KCS_4_CHANNEL			4

#define MAX_KCS_PKT_LEN			(512)

#define INT_ENABLE  			0x1E

/* Different Phases of the KCS Module */
#define KCS_PHASE_IDLE			    0x00
#define KCS_PHASE_WRITE			    0x01
#define KCS_PHASE_WRITE_END		    0x02
#define KCS_PHASE_READ			    0x03
#define KCS_PHASE_ABORT			    0x04
#define KCS_PHASE_ERROR			    0x05
#define ABORT_PHASE_ERROR1		    0x02
#define ABORT_PHASE_ERROR2		    0x03

/* VSC452 Supports four kcs channels, but currently two kcs and BBSP is enabled */
#define MAX_KCS_CHANNELS		   4 

/* KCS Error Codes */
#define KCS_NO_ERROR				0x00
#define KCS_ABORTED_BY_COMMAND		0x01
#define KCS_ILLEGAL_CONTROL_CODE	0x02
#define KCS_LENGTH_ERROR			0x06
#define KCS_UNSPECIFIED_ERROR		0xff

/**** Command Completion Codes ****/
#define	CC_NORMAL					0x00

/* Different KCS States */
#define KCS_IDLE_STATE	            0x00
#define KCS_READ_STATE	            0x40
#define KCS_WRITE_STATE             0x80
#define KCS_ERROR_STATE	            0xC0

/* KCS Command Control Codes */
#define KCS_GET_STATUS				0x60
#define KCS_ABORT					0x60
#define KCS_WRITE_START				0x61
#define KCS_WRITE_END				0x62
#define KCS_READ					0x68

#define KCS_CHANNEL_NOT_IN_USE		0x00
#define KCS_CHANNEL_IN_USE		0x01




/**
 * @def READ_KCS_STATUS_REG
 * @brief Reads a byte from KCS port status register.
 * @param CHANNEL The KCS channel number.
 * @param STATUS a byte data.
 **/
#define READ_KCS_STATUS_REG(CHANNEL, STATUS)                \
do {                                                        \
    switch (CHANNEL) {                                      \
    case KCS_1_CHANNEL: STATUS = *(kcs1_vadd + STATUS_REG); break;      \
    case KCS_2_CHANNEL: STATUS = *(kcs2_vadd + STATUS_REG); break;      \
    case KCS_3_CHANNEL: STATUS = *(kcs3_vadd + STATUS_REG); break;	\
	case KCS_4_CHANNEL: STATUS = *(kcs4_vadd + STATUS_REG); break;		\
    }                                                       \
} while (0)


/*** Local Macro Definitions ***/
/**
 * @def WRITE_KCS_STATUS_REG
 * @brief Writes a byte into KCS port status register.
 * @param CHANNEL The KCS channel number.
 * @param STATUS a byte data.
 **/
#define WRITE_KCS_STATUS_REG(CHANNEL, STATUS)               \
do {                                                        \
    switch (CHANNEL){                                       \
    case KCS_1_CHANNEL: *(kcs1_vadd + STATUS_REG) = STATUS; break;      \
    case KCS_2_CHANNEL: *(kcs2_vadd + STATUS_REG) = STATUS; break;      \
    case KCS_3_CHANNEL: *(kcs3_vadd + STATUS_REG) = STATUS; break;      \
	case KCS_4_CHANNEL: *(kcs4_vadd + STATUS_REG) = STATUS; break;		\
   }                                                       \
} while (0)




/** 
 * @def IS_WRITE_TO_CMD_REG
 * @brief Macro that returns true if last write was to command register.
**/
#define IS_WRITE_TO_CMD_REG(STATUS_REG) (0 != ((STATUS_REG) & 0x08))
#define IBF_BIT_SET(STATUS_REG) (0 != ((STATUS_REG) & 0x02))
#define OBF_BIT_SET(STATUS_REG) (0 != ((STATUS_REG) & 0x01))




/** 
 * @def SET_KCS_STATE
 * @brief Macro to set the current state of the KCS Interface.
**/
#define SET_KCS_STATE(CHANNEL, STATE) 			\
do {											\
	u8 Status = 0;							\
	READ_KCS_STATUS_REG ((CHANNEL), Status);		\
	Status = ((Status & (~0xC0)) | (STATE));		\
	WRITE_KCS_STATUS_REG ((CHANNEL),Status);		\
} while (0)



/**
 * @def KCS_CMD_REG
 * @brief Reads a byte from KCS port command register.
 * @param CHANNEL The KCS channel number.
 * @param CMD a byte data.
 **/
#define KCS_CMD_REG(CHANNEL, CMD)                           \
do {                                                        \
    switch (CHANNEL) {                                      \
    case KCS_1_CHANNEL: CMD = *(kcs1_vadd + CMD_IN); break;              \
    case KCS_2_CHANNEL: CMD = *(kcs2_vadd + CMD_IN); break;              \
    case KCS_3_CHANNEL: CMD = *(kcs3_vadd + CMD_IN); break;              \
	case KCS_4_CHANNEL: CMD = *(kcs4_vadd + CMD_IN); break; 			 \
   }                                                       \
} while (0)


/**
 * @def KCS_DATA_OUT_REG
 * @brief Writes a byte into KCS port data register.
 * @param CHANNEL The KCS channel number.
 * @param DATA a byte data.
 **/
#define KCS_DATA_OUT_REG(CHANNEL, DATA)                     \
do {                                                        \
    switch (CHANNEL) {                                      \
    case KCS_1_CHANNEL: *(kcs1_vadd + DATA_OUT) = DATA; break;             \
    case KCS_2_CHANNEL: *(kcs2_vadd + DATA_OUT) = DATA; break;             \
    case KCS_3_CHANNEL: *(kcs3_vadd + DATA_OUT) = DATA; break;             \
	case KCS_4_CHANNEL: *(kcs4_vadd + DATA_OUT) = DATA; break;			   \
   }                                                       \
} while (0)



/**
 * @def KCS_DATA_IN_REG
 * @brief Reads a byte from KCS port data register.
 * @param CHANNEL The KCS channel number.
 * @param DATA a byte data.
 **/
#define KCS_DATA_IN_REG(CHANNEL, DATA)                      \
do {                                                        \
    switch (CHANNEL) {                                      \
    case KCS_1_CHANNEL: DATA = *(kcs1_vadd + DATA_IN); break;    \
    case KCS_2_CHANNEL: DATA = *(kcs2_vadd + DATA_IN); break;	\
    case KCS_3_CHANNEL: DATA = *(kcs3_vadd + DATA_IN); break;	\
	case KCS_4_CHANNEL: DATA = *(kcs4_vadd + DATA_IN); break;	\
    }                                                       \
} while (0)


/** 
 * @def CLEAR_IBF_STATUS
**/
#define CLEAR_IBF_STATUS(CHANNEL) 			\
do {											\
	u8 Status = 0;							\
	READ_KCS_STATUS_REG ((CHANNEL), Status);		\
	Status = Status & (~0x02);		\
	WRITE_KCS_STATUS_REG ((CHANNEL),Status);		\
} while (0)


/** 
 * @def SET_OBF_STATUS
**/
#define SET_OBF_STATUS(CHANNEL) 			\
do {											\
	u8 Status = 0;							\
	READ_KCS_STATUS_REG ((CHANNEL), Status);		\
	Status = Status | 0x01 ; 	   	\
	WRITE_KCS_STATUS_REG ((CHANNEL),Status);		\
} while (0)



typedef struct 
{
	wait_queue_head_t queue;
	struct timer_list tim;
	volatile u8  Timeout;
} Kcs_OsSleepStruct;


typedef struct
{
	u32		    		KCSRcvPktIx;
	u8*					pKCSRcvPkt;
	u32		    		KCSSendPktIx;          
	u32		    		KCSSendPktLen;       
	u8*					pKCSSendPkt;
	u8		    		KCSPhase;	
	u8		    		AbortPhase;
	u8		    		KCSError;
	u8 					FirstTime;
	Kcs_OsSleepStruct 	KcsReqWait;
	u8					KcsIFActive;
	u8					KcsWakeup;
	u8					KcsResFirstTime;
	u8					KcsINuse;
	Kcs_OsSleepStruct 	KcsResWait;
	u8 					KcsWtIFActive;
	u8 					TxReady;
	u8					PrevCmdAborted;
} KCSBuf_T;


typedef struct
{
	u8			Channel;		/**< Originator's channel number 		*/
	u8			Param;			/**< Parameter							*/
	u8			Cmd;			/**< Command that needs to be processed */
	u8			NetFnLUN;		/**< Net function and LUN of command    */
	u8			Privilege;		/**< Current privilege level			*/
	u32			SessionID;		/**< Session ID if any					*/
	u8			RequestSize;						/**< Size of the Request	*/
	u8			Request	 [MSG_PAYLOAD_SIZE];		/**< Request Data	buffer	*/
	u32			ResponseSize;						/**< Size of the response	*/  //Modified by winston for YAFU
	u8			Response [MSG_PAYLOAD_SIZE];		/**< Response Data buffer	*/
    u8* 	    pRequest;                   /**< Pointer to Request data    */ 
    u8*    		pResponse;                  /**< Pointer to Response data   */

} IPMICmdMsg_T;


/* KCS Request Structure */
typedef struct
{
	u8			NetFnLUN;
	u8			Cmd;

} KCSReq_T;


#define BBSP_FILE  "/dev/kcs3"



#endif /* KCS_H */

