#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/cacheflush.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ide.h>
#include "adc_vsc.h"
#include "adc.h"

#define ADC_MEM_CHECK_REG_FAILS		0xBB
#define ADC_REQ_MEM_REG_FAILS		0xCC
#define ADC_MEM_REMAP_FAILS			0x0A
#define ADC_REG_CHR_DEV_FAILS 		0xAA
#define SET_ADC_OPTION				4



/* ------------------------------------------------------------------------- *
 *  Function prototypes
 * ------------------------------------------------------------------------- */
static int adc_vsc_int(void );
static void  adc_vsc_exit(void );
static int adc_module_open(struct inode *inode, struct file *file);
static int adc_module_close(struct inode *inode, struct file *file);
static int adc_module_ioctl(struct inode *inode, struct file *file, uint cmd, ulong arg);
extern int write_init_st(int option,int value);

/* ------------------------------------------------------------------------- 
 *  Static variables 
 * ------------------------------------------------------------------------- */
static struct file_operations adc_module_ops;
static void *adc_remap_addr ;
static DECLARE_WAIT_QUEUE_HEAD(wq);
static short init_adc_status=0;

/* ADC Driver Private data */
typedef struct {

	int (*read_channel)  ( void * adc_base_addr, uint16_t * adc_value, int channel );
	int (*get_reference_voltage) ( void *adc_base_addr, uint16_t *adc_resolution );
	int (*get_resolution)    ( void *adc_base_addr, uint16_t *adc_ref_volatge);
	int (*enable_ext_ref_voltage)(void *adc_base_addr);
	int  (*disable_ext_ref_voltage)(void *adc_base_addr);	
	
   }  __attribute__((packed)) adc_drv_t; 


/* ------------------------------------------------------------------------- 
 *  Functions
 * ------------------------------------------------------------------------- */

/*** iowrite ***/
inline 
void adc_write_reg(unsigned long addr, unsigned long val)
{
    *((volatile uint32_t*)addr) = val;
}


/*** ioread ***/
inline
unsigned long adc_read_reg(unsigned long addr )
{
    return ((*(volatile uint32_t*)addr));
}


static int
adc_vsc_int(void )
{
	int err,ret=0;
	uint32_t adc_base;

	memset(&adc_module_ops,0,sizeof(adc_module_ops));
	adc_module_ops.open=adc_module_open;
	adc_module_ops.release=adc_module_close;
	adc_module_ops.ioctl= adc_module_ioctl;
	adc_module_ops.read=NULL;
	adc_module_ops.write=NULL;
	adc_module_ops.owner=THIS_MODULE;

    /* ----  Register the character device ------------------- */
	if ((err = register_chrdev(ADC_MAJOR_NUM,DRIVER_NAME,&adc_module_ops)) < 0) 
	{
		printk (KERN_ERR "adc_vsc.c: ADC register_chrdrv failed. err = %d\n", err);
		init_adc_status=ADC_REG_CHR_DEV_FAILS;
		return -ENODEV;
	}

	/* Remap registers */
	 adc_base= ADC_BASE_OFFSET;
	if ((adc_remap_addr = ioremap_nocache(adc_base, ADC_REG_SIZE)) == NULL) 
	{
		printk(KERN_ERR "failed to map GPIO  IO space %08x-%08x to memory",adc_base, adc_base + ADC_REG_SIZE - 1);
		init_adc_status=ADC_MEM_REMAP_FAILS;
		ret = -EIO;
		goto fail;
	}

	/* check gpio memory region */
	if ((err = check_mem_region(adc_base, ADC_REG_SIZE)) < 0)  
	{
		printk(KERN_ERR "ADC IO space %08x-%08x already in use (err %d)",adc_base, adc_base + ADC_REG_SIZE - 1, err);
		init_adc_status=ADC_MEM_CHECK_REG_FAILS;
		ret = -EBUSY;
		goto fail;
	}

	/* request for the memory region */	
	if (request_mem_region(adc_base,ADC_REG_SIZE,DRIVER_NAME)  == NULL) 
	{ 
		printk(KERN_ERR "IO space %08x-%08x adc request mem region failed (err %d)",adc_base, adc_base + ADC_REG_SIZE - 1, err);
		init_adc_status=ADC_REQ_MEM_REG_FAILS;
		ret = -ENODEV;
		goto fail;
	}

	/* Initialize adc here */
    init_adc(adc_remap_addr);
	write_init_st(SET_ADC_OPTION,init_adc_status);	
	printk(KERN_ALERT "ADC Module was  sucessully loaded \n");
	return ret;
    
fail:
	adc_vsc_exit();
	return ret;
}


void 
adc_vsc_exit(void )
{
	uint32_t adc_base;

	printk (KERN_ALERT "Unloading VSC452 ADC driver ..\n");
	adc_base=ADC_BASE_OFFSET;
	exit_adc(adc_remap_addr);
	
	if(init_adc_status==0)
	{
		init_adc_status=0xff;
	}
	write_init_st(SET_ADC_OPTION,init_adc_status);

	/* release the memory occupied */  
	release_mem_region(adc_base,ADC_REG_SIZE);

	/* Unmap registers */
	iounmap( adc_remap_addr); 
	/* ----  UnRegister the character device ------------------- */
	unregister_chrdev(ADC_MAJOR_NUM,DRIVER_NAME);

	return ;

}


void
init_adc(void *adc_base_addr)
{
/* ADC power-on sequence */ 

	/* Enable external reference voltage */
	SET_ADC_BIT(ADCPC,2);

	/* Power-up ADC and release Reset */
	adc_write_reg((uint32_t)ADCPC,(uint32_t)0x00000000);

	/* Assert-Reset and Maintain ADC Power_ON state */  
	adc_write_reg((uint32_t)ADCPC,(uint32_t)0x00000002);

	/* Deassert-Reset and Maintain ADC Power_ON state */
	adc_write_reg((uint32_t)ADCPC,(uint32_t)0x00000000);

	/* Enable external reference voltage */
	SET_ADC_BIT(ADCPC,2);

	return;
}
 

void 
exit_adc(void *adc_base_addr)
{
	/* ADC power-off */ 
	adc_write_reg((uint32_t)ADCC,(uint32_t) 0x000000000);  
	adc_write_reg((uint32_t)ADCPC,(uint32_t)0x00000003);

	return;
}


int 
adc_module_open(struct inode *inode, struct file *file)
{
	adc_drv_t *adc_drv  = (adc_drv_t*) kmalloc ( sizeof(adc_drv_t), GFP_KERNEL );
	memset(adc_drv, 0, sizeof(adc_drv_t));

	/* Initialize the private data */	
	adc_drv->read_channel=adc_read_channel;
	adc_drv->get_reference_voltage=adc_get_reference_voltage;
	adc_drv->get_resolution=adc_get_resolution;
	adc_drv->enable_ext_ref_voltage=enable_external_reference_volt;
	adc_drv->disable_ext_ref_voltage=disable_external_reference_volt;
	file->private_data = adc_drv;

	/* Set Automatic Convertion Enable ADC */  
	adc_write_reg((uint32_t)ADCC,(uint32_t)0x00000100);

	/* Wait until the conversion is complete for all the channels , ie ICC (Initial Conversion Complete) bit is set */  
	wait_event_interruptible_timeout(wq, 1==(((*ADCCS)&0x00000002)>>1),0.25*HZ);

	return 0;
}


int 
adc_module_close(struct inode *inode, struct file *file)
{
	adc_drv_t *adc_drv;
	adc_drv = file->private_data;
	file->private_data = NULL;
	kfree(adc_drv);

	/* Disable Automatic Convertion Flag */  
	adc_write_reg((uint32_t)ADCC,(uint32_t)0x00000000); 

	return 0;
}


int 
adc_module_ioctl(struct inode *inode, struct file *file, uint cmd, ulong arg)
{
	adc_drv_t *adc_drv;
	get_adc_value_t  get_adc_value;
	uint16_t  ref_voltage ,resolution; 

	adc_drv =file->private_data; 

	switch (cmd)
	{
		case READ_ADC_CHANNEL :

			copy_from_user ((void*)&get_adc_value, (void *)arg,sizeof (get_adc_value_t));
			if (adc_drv->read_channel)
			{
				if (0 != adc_drv->read_channel(adc_remap_addr,&get_adc_value.channel_value,get_adc_value.channel_num ))
				return -1;
			}
			copy_to_user ((void*)arg, (void*)&get_adc_value, sizeof (get_adc_value_t));
			break;

		case READ_ADC_REF_VOLATGE:

			if (adc_drv->get_reference_voltage)
			{
				if (0 != adc_drv->get_reference_voltage(adc_remap_addr,&ref_voltage ))
				return -1;
			}
			copy_to_user ((void*)arg, (void*)&ref_voltage, sizeof (uint16_t));
			break;

		case  READ_ADC_RESOLUTION:

			if (adc_drv->get_resolution)
			{
				if (0 != adc_drv->get_resolution(adc_remap_addr,&resolution ))
				return -1;
			}
			copy_to_user ((void*)arg, (void*)&resolution, sizeof (uint16_t));
			break;
			
		case  ENABLE_EXT_REF_VOLTAGE:
			
			if (adc_drv->enable_ext_ref_voltage)
			{
				if (0 != adc_drv->enable_ext_ref_voltage(adc_remap_addr))
				return -1;
			}
			break;

		case  DISABLE_EXT_REF_VOLTAGE:

                  	if (adc_drv->disable_ext_ref_voltage)
			{
				if (0 != adc_drv->disable_ext_ref_voltage(adc_remap_addr))
				return -1;
			}
			break;

		default :
			printk (KERN_ALERT "Invalid ADC driver ioctl\n");
			return -EINVAL;
		}
	
	return 0;
}


module_init(adc_vsc_int);
module_exit(adc_vsc_exit);


int 
adc_read_channel(void *adc_base_addr,  uint16_t*adc_value, int channel)
{
	volatile uint32_t rawdata =0;
		
	/*Check ADC Channel  is between 0 -11 */
	if ((channel < ADC_CHANNEL0_NUM) || (channel > ADC_CHANNEL12_NUM))            
	{
	printk (KERN_ERR "Invalid ADC Channel Number = 0x%x\n", channel);
	return -1;
	}

	/*Get raw data from the ADC channel*/
	rawdata=(uint32_t)adc_read_reg((uint32_t)ADCCD(channel)); 

	/*Typecast the rawdata (32bit) value to 16 bit User API varibale */  
	*adc_value=(uint16_t)rawdata;

	return 0;
}
 

int 
adc_get_resolution(void *adc_base_addr, uint16_t *adc_resolution)
{
	*adc_resolution = ADC_NUM_BITS_FOR_RESOLUTION;
	return 0;
}
 

int 
adc_get_reference_voltage(void *adc_base_addr, uint16_t *adc_ref_volatge)
{
	volatile uint32_t reg_buff =0;
	reg_buff=(uint32_t)adc_read_reg((uint32_t)ADCPC);
	reg_buff=((reg_buff & (uint32_t)0x00000004))>>2;

	if (1==reg_buff)
	{
		*adc_ref_volatge = ADC_EXTERNAL_REF_VOLTAGE_IN_MVOLTS;
	}
	else
	{
		*adc_ref_volatge = ADC_INTERNAL_REF_VOLTAGE_IN_MVOLTS;
	}

	return 0;
}


int 
enable_external_reference_volt(void *adc_base_addr)
{
	SET_ADC_BIT(ADCPC,2);
	return 0;
}


int
disable_external_reference_volt(void *adc_base_addr)
{
	CLEAR_ADC_BIT(ADCPC,2);
	return 0;
}

 
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("VSC452 ADC Driver");


