/*
 * NAND flash support on VSC452
 *
 * Copyright (C) 2008 American Megatrends Inc
 *
 */

#include <linux/init.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/delay.h>

#ifdef CONFIG_MTD_PARTITIONS
extern int CreateNandPartitions(struct mtd_partition *part,unsigned long nandsize);
#define MAX_PARTITIONS 16
#else
#define MAX_PARTITIONS 1
#endif

static struct mtd_info *vsc_mtd  = NULL;


/* driver internal data */
struct vsc_local {
	struct mtd_partition part[MAX_PARTITIONS];
	int hwctl;
};


//#define FULL_FLASH_PARTITION_NAME	"NAND-fullpart"
//static char *fullpart = FULL_FLASH_PARTITION_NAME;



#define VSC_NAND_BASE			(0xB0200040)

#define NAND_RD_DATA(base)		*((volatile unsigned long *)((base)+0x00))
#define NAND_WR_DATA(base)		*((volatile unsigned long *)((base)+0x04))
#define NAND_ADDR(base)			*((volatile unsigned long *)((base)+0x08))
#define NAND_COMMAND(base)		*((volatile unsigned long *)((base)+0x0C))
#define NAND_CONTROL(base)		*((volatile unsigned long *)((base)+0x10))
#define NAND_STATUS(base)		*((volatile unsigned long *)((base)+0x14))

#define STATUS_READY			0x02


/* VSC452 does not have support to control signals. It does 
   automatically the command and address register are written */
#define COMMAND_PHASE 	  0x01
#define ADDR_PHASE	  0x02
static void  nand_write_byte(struct mtd_info *mtdinfo, u_char byte);

static void 
nand_hwcontrol(struct mtd_info *mtdinfo, int cmd, unsigned int ctrl)
{
	struct nand_chip *this = mtdinfo->priv;
	struct vsc_local *local = this->priv;

	if( ctrl & NAND_CTRL_CHANGE ) 
	 {
		if( ctrl & NAND_ALE )
			local->hwctl |= ADDR_PHASE;
		else  
                       local->hwctl &= ~ADDR_PHASE;
		
		if ( ctrl & NAND_CLE )	
		        local->hwctl |= COMMAND_PHASE;	
               else 
		        local->hwctl &= ~COMMAND_PHASE;
 #if 0		 	
		 /* we not handling this */  	
               if ( ctrl & NAND_NCE ) 
		        //nand_cs_on(); 
		else 	
 		        //nand_cs_off();   	
 #endif          
	  }
		
	
	if (cmd != NAND_CMD_NONE)
		nand_write_byte(mtdinfo, cmd);
	return;
}

static void 
nand_write_byte(struct mtd_info *mtdinfo, u_char byte)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_W; 
	struct vsc_local *local = this->priv;

	if (local->hwctl & COMMAND_PHASE)
		NAND_COMMAND(base) = byte;
	else 
	{
		if (local->hwctl & ADDR_PHASE)
			NAND_ADDR(base) = byte;
		else
			NAND_WR_DATA(base) = byte;
	}
}

static u_char 
nand_read_byte(struct mtd_info *mtdinfo)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	return NAND_RD_DATA(base);
}

static int
nand_dev_ready(struct mtd_info *mtdinfo)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	if  (!(NAND_STATUS(base) & STATUS_READY))
		return 0;
	return 1;
}

static void 
nand_read_buf(struct mtd_info *mtdinfo, uint8_t *buf, int len)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	for (;len > 0; len --)
		*buf++ = NAND_RD_DATA(base);
	return;
}

static void 
nand_write_buf(struct mtd_info *mtdinfo, const uint8_t *buf, int len)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_W;

	for (; len > 0; len --)
			NAND_WR_DATA(base) =  *buf++;
}

static int 
nand_verify_buf(struct mtd_info *mtdinfo, const uint8_t *buf, int len)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	for (; len > 0; len --)
		if (*buf++ !=NAND_RD_DATA(base))
			return -1;

	return 0;
}


static int __init
init_vsc_nand(struct mtd_info *mtdinfo)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	/* Enable Chip Select CS0 */
	NAND_CONTROL(base) = 0x10;
	return 0;
}

static void __exit
exit_vsc_nand(struct mtd_info *mtdinfo)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	/* Disable Chip Select CS0 */
	NAND_CONTROL(base) = 0;
	return;
}

static int __init 
vsc_nand_init_module(void)
{
	struct nand_chip *chip;
	struct vsc_local *local;
	int allocsize;
	int retval = -ENXIO;
	unsigned long nandsize;
	int numparts = 1;


	/* MTD Information, Nand Chip Informatiion and Local data */
	allocsize = sizeof(struct mtd_info) + sizeof(struct nand_chip) + sizeof (struct vsc_local);

	/* Allocate memory for MTD Structures */
	vsc_mtd = kmalloc(allocsize ,GFP_KERNEL);
	if (!vsc_mtd) 
	{
		printk(KERN_ERR "vsc_nand:Unable to allocate MTD structures\n");
		return -ENOMEM;
	}
	memset(vsc_mtd, 0, allocsize );
	chip         = (struct nand_chip *)(vsc_mtd + 1); /* Nand Chip info is next to mtd info */
	local         = (struct vsc_local *)(chip + 1);	  /* Local data is next to Nand Chip info */
        vsc_mtd->priv = (void *)chip;
	chip->priv   = (void *)local;

	
	/* Register hardware functions */
	chip->cmd_ctrl   = nand_hwcontrol;
	chip->read_byte  = nand_read_byte;
	chip->dev_ready  = nand_dev_ready;
//	chip->write_byte = nand_write_byte;
	chip->write_buf  = nand_write_buf;
	chip->read_buf   = nand_read_buf;
	chip->verify_buf = nand_verify_buf;
	chip->ecc.mode    = NAND_ECC_SOFT;
	chip->IO_ADDR_R  = chip->IO_ADDR_W = (void *)VSC_NAND_BASE;

	/* Do any init for local data structures - driver and hardware specific */
	local->hwctl = 0;
	
	/* Register driver */
	retval = nand_scan(vsc_mtd, 1);
	if (retval!= 0) 
	{
		printk(KERN_ERR"vsc_nand: Can't register VSC NAND driver\n");
		goto error;
	}

	retval = init_vsc_nand(vsc_mtd);
	if (retval != 0)
	{
		printk(KERN_ERR"vsc_nand: Can't initialize VSC NAND hardware\n");
		goto error;
	}

	nandsize = vsc_mtd->size;

	/* Register whole NAND as one big partition */
//	local->part[0].name 	= fullpart;
//	local->part[0].offset 	= 0;
//	local->part[0].size 	= nandsize;

#ifdef CONFIG_MTD_PARTITIONS
	numparts = CreateNandPartitions(local->part,nandsize);
#endif

//	add_mtd_partitions(vsc_mtd, &local->part[0], numparts+1); 
	add_mtd_partitions(vsc_mtd, &local->part[0], numparts); 
        return 0;

error:
	kfree(vsc_mtd);

	if (retval > 0)
		retval = -ENXIO;
	return retval;
}


static void __exit 
vsc_nand_exit_module(void)
{
	exit_vsc_nand(vsc_mtd);  	/* Disable Hardware */
	nand_release(vsc_mtd); 		/* Unregister driver */
	kfree(vsc_mtd);        		/* Free all structures */
}

module_init(vsc_nand_init_module);
module_exit(vsc_nand_exit_module);

MODULE_LICENSE ("GPL");
MODULE_AUTHOR ("American Megatrends Inc");
