/*
 * Tiny TTY driver
 *
 * Copyright (C) 2002 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 * This driver shows how to create a minimal serial driver using the 2.5
 * kernel's serial driver layer. It does not rely on any backing hardware, but
 * creates a timer that emulates data being received from some kind of
 * hardware.
 *
 * Compiled this driver with:
 
	CFLAGS = -Wall -O2 -fomit-frame-pointer -DMODULE -D__KERNEL__
	IDIR = /path/to/linux-2.5/include/goes/here

	tiny_serial.o::tiny_serial.c
		$(CC) $(CFLAGS) -I$(IDIR) -c -o $@ $<
 */




#if defined(CONFIG_SERIAL_VSC452_UART_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
#define SUPPORT_SYSRQ
#endif

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/mca.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial_reg.h>
#include <linux/serial_core.h>
#include <linux/serial.h>
#include <linux/serial_8250.h>
#include <linux/nmi.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm/mach-vsc452/vsc452.h>
#include <asm/addrspace.h>


/* Hardware interrupt forcing register */
#define HIF_REG         (volatile unsigned long *)0xB02008f4



/* Interrupt observe registers */
#define IOR0_REG        (volatile unsigned long *)0xB02008f8
#define IOR1_REG        (volatile unsigned long *)0xB02008fc


/*
 * Configuration:
 *   share_irqs - whether we pass IRQF_SHARED to request_irq().  This option
 *                is unsafe when used on edge-triggered interrupts.
 */
static unsigned int share_irqs = 1;

/*
 * Debugging.
 */

#if 0
#define DEBUG_INTR(fmt...)	printk(fmt)
#else
#define DEBUG_INTR(fmt...)	do { } while (0)
#endif

#define PASS_LIMIT	256

#include <asm/serial.h>

#define UART_NR	2 



struct uart_8250_port {
	struct uart_port	port;
	struct list_head	list;		/* ports on this IRQ */
	unsigned char		tx_ier;
	unsigned char		rx_ier;
};

static struct uart_8250_port serial8250_ports[UART_NR];


struct timer_list dbg_timer;	 

int dbgflag =0;

static unsigned int ser_rr(unsigned int offset, struct uart_8250_port * port)
{
	int portnum = port->port.line;	
	
//	return *(unsigned long*)(serial8250_ports[portnum].port.membase + offset );
	  
	return *(unsigned long*)(serial8250_ports[portnum].port.membase +( offset - SYS_GPURS0) );

}

static void ser_wr(unsigned int offset, unsigned int value,struct uart_8250_port *port)
{
	int portnum = port->port.line;
//	*(unsigned long*)(serial8250_ports[portnum].port.membase + offset ) = value;

	*(unsigned long*)(serial8250_ports[portnum].port.membase +( offset - SYS_GPURS0) ) = value;

   
}



static inline void __stop_tx(struct uart_8250_port *up)
{
	if(up->tx_ier & SYS_GPUIEx_TRIE)
	{
		up->tx_ier &= ~SYS_GPUIEx_TRIE;
		ser_wr(SYS_GPUIE0,up->tx_ier,up);
	}
}

static void serial8250_stop_tx(struct uart_port *port)
{
	struct uart_8250_port *up = (struct uart_8250_port *)port;

	//printk(KERN_ERR "serial8250_stop_tx\n");

	__stop_tx(up);
}

static void transmit_chars(struct uart_8250_port *up);

static void serial8250_start_tx(struct uart_port *port)
{
	struct uart_8250_port *up = (struct uart_8250_port *)port;

//	printk(KERN_ERR "serial8250_start_tx\n");

	if(! (up->tx_ier & SYS_GPUIEx_TRIE) )
	{
		up->tx_ier |= SYS_GPUIEx_TRIE;
		// enable interrupt generation
		ser_wr(SYS_GPUIE0, up->tx_ier,up);
		// kick off the first tx, which will cause an int on tx
		transmit_chars(up);
	}
}

static void serial8250_stop_rx(struct uart_port *port)
{
	struct uart_8250_port *up = (struct uart_8250_port *)port;

	//printk(KERN_ERR "serial8250_stop_rx\n");

	//todo disable ints

	up->rx_ier &= ~SYS_GPURSx_RXR;
	ser_wr(SYS_GPURS0, up->rx_ier,up);

}

static void serial8250_enable_ms(struct uart_port *port)
{
	//printk(KERN_ERR "serial8250_enable_ms\n");
}

#if 0
static int printregs(void)
{
	printk ( "****** UART0 regs ************\n");
	printk ( " %x \n",ser_rr(SYS_GPURS0,&serial8250_ports[0]));
	printk ( " %x \n",ser_rr(SYS_GPURD0,&serial8250_ports[0]));
	printk ( " %x \n",ser_rr(SYS_GPUTS0,&serial8250_ports[0]));
	printk ( " %x \n",ser_rr(SYS_GPUTD0,&serial8250_ports[0]));
	printk ( " %x \n",ser_rr(SYS_GPUBR0,&serial8250_ports[0]));
	printk ( " %x \n",ser_rr(SYS_GPUIS0,&serial8250_ports[0]));
	printk ( " %x \n",ser_rr(SYS_GPUIE0,&serial8250_ports[0]));

	printk ( "****** UART1 regs ************\n");
	printk ( " %x \n",ser_rr(SYS_GPURS0, &serial8250_ports[1]));
	printk ( " %x \n",ser_rr(SYS_GPURD0, &serial8250_ports[1]));
	printk ( " %x \n",ser_rr(SYS_GPUTS0, &serial8250_ports[1]));
	printk ( " %x \n",ser_rr(SYS_GPUTD0, &serial8250_ports[1]));
	printk ( " %x \n",ser_rr(SYS_GPUBR0, &serial8250_ports[1]));
	printk ( " %x \n",ser_rr(SYS_GPUIS0, &serial8250_ports[1]));
	printk ( " %x \n",ser_rr(SYS_GPUIE0, &serial8250_ports[1]));
		
}
#endif
#define SET_BIT(addr,bitmask)   (addr|bitmask)
#define CLEARBIT(addr,bitmask)  (addr&bitmask)

static void receive_chars(struct uart_8250_port *up, int *status)
{
	struct tty_struct *tty = up->port.info->port.tty;
	unsigned char ch;
	unsigned int intrstatus = 0;

	intrstatus = ser_rr(SYS_GPURS0, up);
	if ( 0x00000080 & intrstatus )
	{
		ch = (unsigned char)ser_rr(SYS_GPURD0,up);
		up->port.icount.rx++;
		tty_insert_flip_char(tty, ch, TTY_NORMAL);
	}

	/* reset interrupt*/
	*(unsigned long *)(up->port.membase + (SYS_GPUIS0-SYS_GPURS0)) |= 0x04;
	
	tty_flip_buffer_push(tty);
}

static unsigned int serial8250_tx_empty(struct uart_port *port);

static void transmit_chars(struct uart_8250_port *up)
{
	struct circ_buf *xmit = &up->port.info->xmit;
	unsigned int intr;
	if (uart_circ_empty(xmit))
	{
		__stop_tx(up);
		//clear the transmit ready intr
		intr = ser_rr(SYS_GPUIS0,up);
		*(unsigned long*)(up->port.membase + ( SYS_GPUIS0-SYS_GPURS0))|=0x01;
 		return;
	}



	
	if ( ser_rr(SYS_GPUTS0,up) & 0x01)
	{
		ser_wr(SYS_GPUTD0, (unsigned int)xmit->buf[xmit->tail],up);
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		*(unsigned long*)(up->port.membase + ( SYS_GPUIS0-SYS_GPURS0))|=0x01;
		up->port.icount.tx++;
	}


	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&up->port);



}


#define UART0INTR_MASK   0x00010000
#define UART1INTR_MASK	 0x00020000

/*
 * This is the serial driver's interrupt routine.
 *
 * Arjan thinks the old way was overly complex, so it got simplified.
 * Alan disagrees, saying that need the complexity to handle the weird
 * nature of ISA shared interrupts.  (This is a special exception.)
 *
 * In order to handle ISA shared interrupts properly, we need to check
 * that all ports have been serviced, and therefore the ISA interrupt
 * line has been de-asserted.
 *
 * This means we need to loop through all ports. checking that they
 * don't have an interrupt pending.
 */
static irqreturn_t serial8250_interrupt(int irq, void *dev_id)
{
//	struct uart_8250_port *up = &serial8250_ports[0];
	struct uart_8250_port *up = dev_id;
	unsigned int status;
	unsigned int  ior0_val,ior1_val;
	unsigned long flags = 0;
	int temp = 0;

	DEBUG_INTR("serial8250_interrupt(%d)...", irq);

	
	/* Read the interrupt status registers */
	 spin_lock_irqsave(&up->port.lock, flags);
	 ior0_val = *IOR0_REG;
	 ior1_val = *IOR1_REG;
	 *HIF_REG = 0; /* thanks Bala this worked! */
	 spin_unlock_irqrestore(&up->port.lock, flags);

	/*Check whether its uart interrupt*/
	switch ( up->port.line)
	{
		case 0:
			if ( 0 == (ior0_val & UART0INTR_MASK)) 
			{
				return IRQ_NONE;
			}
			break;		
		case 1:
			if( 0 == (ior0_val & UART1INTR_MASK))
			{
				return IRQ_NONE;
			}
			break;
		default:
			return IRQ_NONE;
	}
	

	 
//	spin_lock_irqsave(&up->port.lock,flags);
	
	
	//unsigned int status = serial_inp(up, UART_LSR);
//	status = ser_rr(SYS_GPURS0,up);
	status = ser_rr(SYS_GPUIS0,up);

	if (( status & SYS_GPUISx_GPIS )&& (ser_rr(SYS_GPURS0,up)&SYS_GPURSx_RXR ))
	{
		receive_chars(up, &temp);
	}

	
	if (( status & SYS_GPUISx_TRIS )&& (ser_rr(SYS_GPUTS0,up)&SYS_GPUTSx_TXR ))
	{
//		if ( 1 == up->port.line )
		//	printk("*");
//		printregs();
		transmit_chars(up);
	}

	if ( 0 == (status & (SYS_GPUISx_GPIS | SYS_GPUISx_TRIS)))
	{
		if (ser_rr(SYS_GPURS0,up)&SYS_GPURSx_RXR)
		{
			receive_chars(up, &temp);
		}
		if (ser_rr(SYS_GPUTS0,up)&SYS_GPUTSx_TXR)
		{
	//		if ( 1 == up->port.line )
//				printk("*");

			transmit_chars(up);
		}
	}
//	spin_unlock_irqrestore(&up->port.lock,flags);

#if 0
	DEBUG_INTR("status = %x...", status);
	
	if (status & SYS_GPURSx_RXR)
	{
		receive_chars(up, &status, regs);
		if ( up->port.line == 1 )
		printk ( "intr 1\n");
	
	}	

	status = ser_rr(SYS_GPUTS0,up);

	if (status & SYS_GPUTSx_TXR)
	{
		// clear interrupt
		ser_wr(SYS_GPUIS0, SYS_GPUISx_TRIS,up);
		transmit_chars(up);
	//	printk ( "intr 1\n");
	}
		

//	spin_unlock(&up->port.lock);


	DEBUG_INTR("end.\n");
#endif
	return IRQ_RETVAL(1);
}


static unsigned int serial8250_tx_empty(struct uart_port *port)
{
	struct uart_8250_port *up = (struct uart_8250_port *)port;
	unsigned long flags;
	unsigned int ret;

   	// printk(KERN_ERR "serial8250_tx_empty\n");

	spin_lock_irqsave(&up->port.lock, flags);

	//ret = (ser_rr(SYS_GPUTS0) & SYS_GPUTSx_TXR) ? TIOCSER_TEMT : 0;
	ret = TIOCSER_TEMT;

	spin_unlock_irqrestore(&up->port.lock, flags);

	return ret;
}

static unsigned int serial8250_get_mctrl(struct uart_port *port)
{
	//printk(KERN_ERR "serial8250_get_mctrl\n");
	return TIOCM_DTR | TIOCM_RTS | TIOCM_DSR;
}

static  void serial8250_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
	//printk(KERN_ERR "serial8250_set_mctrl\n");
}

static void serial8250_break_ctl(struct uart_port *port, int break_state)
{
	//printk(KERN_ERR "serial8250_break_ctl\n");
}


static void serial8250_start_rx(struct uart_8250_port *up)
{
	unsigned long flags;
	unsigned int ctrl;

	local_irq_save(flags);
	ctrl = ser_rr(SYS_GPUBR0,up);
	ctrl &= 0xc7 ;
	// RX FIFO disable
	ser_wr(SYS_GPUBR0,ctrl,up);
	// RX FIFO enable and set interrupt on rising edge
//	ctrl |= 0x20;
	ctrl |= 0x08;
	ser_wr(SYS_GPUBR0,ctrl,up);
	ser_wr(SYS_GPUTS0,0,up);
	local_irq_restore(flags);
}

#if 0
static void dbg_timer_func(unsigned long priv)
{
	struct uart_8250_port *up = (struct uart_8250_port *)priv;
	struct tty_struct *tty = up->port.info->tty;
	unsigned char ch;

//	printk(KERN_ERR "receive_chars\n");

	printk ("port0RS %x",ser_rr1(SYS_GPURS0,0));
	printk ("port1RS %x",ser_rr1(SYS_GPURS0,1));
	 
	//ch = serial_inp(up, UART_RX);
	ch = (unsigned char)ser_rr(SYS_GPURD0,up);
	printk ("%c",ch);
	up->port.icount.rx++;

	tty_insert_flip_char(tty, ch, TTY_NORMAL);

//	spin_unlock(&up->port.lock);
	tty_flip_buffer_push(tty);
//	spin_lock(&up->port.lock);

	quit:
	 /* Rearm the timer. */
    dbg_timer.expires = jiffies + (1*HZ);
    add_timer(&dbg_timer);

}

#endif
static int serial8250_startup(struct uart_port *port)
{
	struct uart_8250_port *up = (struct uart_8250_port *)port;
//	int retval = 0;
//	int baud = 0;
//	int ier =0;
//   unsigned int flags; 
//	printk(KERN_ERR "serial8250_startup %x\n",port->line);

	/*
	 * Clear the interrupt registers.
	 */
	(void) ser_rr(SYS_GPURD0,up);

#if 0
	retval = request_irq(up->port.irq, serial8250_interrupt, IRQF_SHARED, "serial", up);

	//HACK: dropping this section because Linux closes this driver and the reopen fails. Need to support close
	if (retval)
	{
		printk(KERN_ERR "serial8250_startup: request_irq failed\n");
		return -ENODEV;
	}
#endif	
	serial8250_start_rx(up);
	
//	spin_lock_irqsave(&up->port.lock, flags);
	

	//set baud rate
//	baud = ser_rr(SYS_GPUBR0,up);
	//ser_wr(SYS_GPUBR0, 0x6A , up);
//	baud = ser_rr(SYS_GPUBR0,up);
		
//	printk ("Starup done %x\n",baud);
    
	// Enable rx interrupts
	up->rx_ier |= SYS_GPURSx_RXR;
	ser_wr(SYS_GPURS0, up->rx_ier,up);
	if ( up->port.line == 1 )
	{
		dbgflag = 1;
	
	}
//	printregs();

#if 0
	if ( up->port.line == 1 )
	{
		/* Setup link check timer */
		init_timer(&dbg_timer);
		dbg_timer.expires = jiffies + (1*HZ);
		dbg_timer.data = (unsigned long)up;
		dbg_timer.function = dbg_timer_func;
		add_timer(&dbg_timer);
	}	
#endif	
//	spin_unlock_irqrestore(&up->port.lock, flags);
	return 0;
}

static void 
serial8250_shutdown(struct uart_port *port)
{
	struct uart_8250_port *up = (struct uart_8250_port *)port;
	unsigned long flags;

	local_irq_save(flags);
	serial8250_stop_tx(&up->port);
	serial8250_stop_rx(&up->port);
	//free_irq(up->port.irq, up);
	/* disbale GPUART interrupt */
//	clear_gpuart_bit (&vport->port, GPURSX, U_RXR);
	local_irq_restore(flags);

	// del_timer_sync(&dbg_timer);
	//TODO:// re-enable irq for SH_INT, This need to remove if new kernel interrupt ready
	/*if (port->line) {
		disable_irq(6);
		enable_irq(6);
	}*/

	//printk(KERN_ERR "serial8250_shutdown: not gonna do it\n");
}

static void
serial8250_set_termios(struct uart_port *port, struct ktermios *termios,
		       struct ktermios *old)
{
#if 1
	unsigned long flags, value ;
	unsigned int baud = 0;
	unsigned int baudval =0;
	struct tty_struct *tty = port->info->port.tty;

	baud = tty_get_baud_rate(tty);
//	printk ( "baud rate %d\n",baud);
//	if ( port->line == 0 )
//		baud = 38400;

	switch ( baud )
	{
		case 9600: 
			baudval =SYS_SIBR_9600;	
			break;
		case 19200: 
			baudval =SYS_SIBR_19200;
			break;
		case 38400: 
			baudval =SYS_SIBR_38400;
			break;
		case 57600:  
			baudval =SYS_SIBR_57600;
			break;
		case 115200: 
			baudval =SYS_SIBR_115200;
			break;
		case 230400:  
			baudval =SYS_SIBR_230400;
			break;

	
	} 
	//printk(KERN_ERR "serial8250_set_termios\n");
	 spin_lock_irqsave(&port->lock,flags);
	  value = *(unsigned int*)(serial8250_ports[port->line].port.membase + (SYS_GPUBR0 - SYS_GPUART0_BASE ) );
	  value = ((value & 0xFFFFFFF8) | baudval);
	 *(unsigned int*)(serial8250_ports[port->line].port.membase +( SYS_GPUBR0 - SYS_GPUART0_BASE ) ) =  value ;//(0x4a);
	  value = *(unsigned int*)(serial8250_ports[port->line].port.membase + (SYS_GPUBR0 - SYS_GPUART0_BASE ) );
	 spin_unlock_irqrestore(&port->lock, flags);
//	 printk ("baud %x\n",value);
	#endif 
	
}

static void
serial8250_pm(struct uart_port *port, unsigned int state,
	      unsigned int oldstate)
{
	//printk(KERN_ERR "serial8250_pm\n");
}

static void 
serial8250_release_port(struct uart_port *port)
{
	//printk(KERN_ERR "serial8250_release_port\n");
}

static int 
serial8250_request_port(struct uart_port *port)
{
	//printk(KERN_ERR "serial8250_request_port\n");
	return 0;
}

static void 
serial8250_config_port(struct uart_port *port, int flags)
{
	//printk(KERN_ERR "serial8250_config_port: why is this being called?\n");
}

static int
serial8250_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	//printk(KERN_ERR "serial8250_verify_port: not allowing change\n");
	return -EINVAL;
}

static const char * 
serial8250_type(struct uart_port *port)
{
	//printk(KERN_ERR "serial8250_type\n");
	return "452uart";
}

static struct uart_ops serial8250_pops = {
	.tx_empty	= serial8250_tx_empty,
	.set_mctrl	= serial8250_set_mctrl,
	.get_mctrl	= serial8250_get_mctrl,
	.stop_tx	= serial8250_stop_tx,
	.start_tx	= serial8250_start_tx,
	.stop_rx	= serial8250_stop_rx,
	.enable_ms	= serial8250_enable_ms,
	.break_ctl	= serial8250_break_ctl,
	.startup	= serial8250_startup,
	.shutdown	= serial8250_shutdown,
	.set_termios	= serial8250_set_termios,
	.pm		= serial8250_pm,
	.type		= serial8250_type,
	.release_port	= serial8250_release_port,
	.request_port	= serial8250_request_port,
	.config_port	= serial8250_config_port,
	.verify_port	= serial8250_verify_port,
};

static void __init 
serial8250_isa_init_ports(void)
{
	static int first = 1;
	int i;

	if (!first)
		return;
	first = 0;

	for (i = 0; i < UART_NR; i++) {
		struct uart_8250_port *up = (struct uart_8250_port *)&serial8250_ports[i];

		up->port.line = i;
		spin_lock_init(&up->port.lock);

		up->port.ops = &serial8250_pops;

		up->port.type = 1000; //just so we're not "unknown"
		up->port.fifosize = 8 ;   //  1;
		up->port.irq      = MIPSCPU_INT_UART_GPIO + MIPSCPU_INT_BASE;
		up->port.uartclk  = 115200 * 16; //dummy to avoid getting dropped in case serial_core cares
		up->port.flags    = 0;
		up->port.hub6     = 0;
		if ( 0 == up->port.line)
		{
			up->port.mapbase  = SYS_GPUART0_BASE;
			up->port.membase  = (char*)SYS_GPUART0_BASE; 
		}
		else if ( 1 == up->port.line )
		{
			up->port.mapbase = SYS_GPUART1_BASE;
			up->port.membase = (char*)SYS_GPUART1_BASE;
		}
		up->port.iotype   = UPIO_MEM32;
		up->port.regshift = 2;
		/*Disable receive interrupts, it will be enabled later on by core*/
		ser_wr (SYS_GPURS0,0x0,up);
		
	}
}

static int __init
serial8250_register_ports(struct uart_driver *drv, struct device *dev)
{
	int i, retval =0;


	printk(KERN_ERR "serial8250_register_ports\n");

	serial8250_isa_init_ports();

	for (i = 0; i < UART_NR; i++) {
//		struct uart_8250_port *up = &serial8250_ports[i];

		//up->port.dev = dev;
	//	uart_add_one_port(drv, &up->port);
		uart_add_one_port(drv,  &serial8250_ports[i].port);

		
		retval = request_irq(serial8250_ports[i].port.irq, serial8250_interrupt, IRQF_SHARED, "serial", &serial8250_ports[i]);
		
		//HACK: dropping this section because Linux closes this driver and the reopen fails. Need to support close
		if (retval)
		{
			printk(KERN_ERR "serial8250_startup: request_irq failed\n");
			return -ENODEV;
		}
	}
}

#ifdef CONFIG_SERIAL_VSC452_UART_CONSOLE


/*
 *	Print a string to the serial port trying not to disturb
 *	any possible real use of the port...
 *
 *	The console_lock must be held when we get here.
 */
extern void prom_putchar(char);

static void
serial8250_console_write(struct console *co, const char *s, unsigned int count)
{
	struct uart_8250_port *up = &serial8250_ports[co->index];
	unsigned int ier;
	int i;

	touch_nmi_watchdog();
	spin_lock(&up->port.lock); 

	 /* First save the IER then disable the interrupts */
	ier = ser_rr(SYS_GPUIE0,up);	
	ser_wr(SYS_GPUIE0, 0,up);

	/* Write each Character */
	for (i = 0; i < count; i++, s++) 
	{
		prom_putchar(*s);
		if (*s == 10) 
			prom_putchar(13);
	}
	/* Restore the IER */
	ser_wr(SYS_GPUIE0, ier,up);

	spin_unlock(&up->port.lock);
}

static int 
serial8250_console_setup(struct console *co, char *options)
{
	struct uart_port *port;

	/*
	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	if (co->index >= UART_NR)
		co->index = 0;
	port = &serial8250_ports[co->index].port;
	if (!port->iobase && !port->membase)
		return -ENODEV;

	return 0;
}

static struct uart_driver serial8250_reg;
static struct console serial8250_console = 
{
	.name		= "ttyS",
	.write		= serial8250_console_write,
	.device		= uart_console_device,
	.setup		= serial8250_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &serial8250_reg,
};

static int __init 
serial8250_console_init(void)
{
	serial8250_isa_init_ports();
	register_console(&serial8250_console);
	return 0;
}
console_initcall(serial8250_console_init);
#endif



static struct uart_driver serial8250_reg = {
	.owner			= THIS_MODULE,
	.driver_name		= "serial",
	.dev_name		= "ttyS",
	.major			= TTY_MAJOR,
	.minor			= 64,
	.nr			= UART_NR,
	.cons			= &serial8250_console
};



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
/////////////////////////////ISA JUNK//////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
/**
 *	serial8250_suspend_port - suspend one serial port
 *	@line:  serial line number
 *      @level: the level of port suspension, as per uart_suspend_port
 *
 *	Suspend one serial port.
 */
void serial8250_suspend_port(int line)
{
	uart_suspend_port(&serial8250_reg, &serial8250_ports[line].port);
}

/**
 *	serial8250_resume_port - resume one serial port
 *	@line:  serial line number
 *      @level: the level of port resumption, as per uart_resume_port
 *
 *	Resume one serial port.
 */
void serial8250_resume_port(int line)
{
	uart_resume_port(&serial8250_reg, &serial8250_ports[line].port);
}

/*
 * Remove serial ports registered against a platform device.
 */
static int __devexit serial8250_remove(struct platform_device *dev)
{
	//printk(KERN_ERR "serial8250_remove: ISA stuff stubbed\n");
	return 0;
}

#if 0
/*
 * Register a set of serial devices attached to a platform device.  The
 * list is terminated with a zero flags entry, which means we expect
 * all entries to have at least UPF_BOOT_AUTOCONF set.
 */
static int __devinit serial8250_probe(struct platform_device *dev)
{
	//printk(KERN_ERR "serial8250_probe: ISA stuff stubbed\n");
	return 0;
}

static int serial8250_suspend(struct platform_device *dev, pm_message_t state)
{
	//printk(KERN_ERR "serial8250_suspend: ISA stuff stubbed\n");
	return 0;
}

static int serial8250_resume(struct platform_device *dev)
{
	//printk(KERN_ERR "serial8250_resume: ISA stuff stubbed\n");
	return 0;
}
#endif
/*static struct platform_driver serial8250_isa_driver = {
	.probe		= serial8250_probe,
	.remove		= __devexit_p(serial8250_remove),
	.suspend	= serial8250_suspend,
	.resume		= serial8250_resume,
	.driver		= {
		.name	= "serial8250",
	},
};*/

/*
 * This "device" covers _all_ ISA 8250-compatible serial devices listed
 * in the table in include/asm/serial.h
 */
static struct platform_device *serial8250_isa_devs;



/* reboot notifiers */
static int vsc_notify_serial_driver(struct notifier_block *this, unsigned long code, void *unused);


static struct notifier_block vsc_serial_notifier =
{
        .notifier_call = vsc_notify_serial_driver,
};

static int vsc_notify_serial_driver(struct notifier_block *this, unsigned long code, void *unused)
{

//	printk ("serial reboot notifier invoked\n");
        serial8250_shutdown(&(serial8250_ports[0].port));
        serial8250_shutdown(&(serial8250_ports[1].port));

        return NOTIFY_DONE;

}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
////////////////////         Init hooks and module setup          /////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static int __init serial8250_init(void)
{
	int ret;

	printk(KERN_INFO "Serial: 8250/16550 driver $Revision: 1.90 $ "
		"%d ports, IRQ sharing %sabled\n", (int) UART_NR,
		share_irqs ? "en" : "dis");

	ret = uart_register_driver(&serial8250_reg);
	
	if (ret)
		goto out;
#if 0	

	serial8250_isa_devs = platform_device_register_simple("serial8250",
					 PLAT8250_DEV_LEGACY, NULL, 0);
	if (IS_ERR(serial8250_isa_devs)) {
		ret = PTR_ERR(serial8250_isa_devs);
		goto unreg;
	}
#endif	

	serial8250_register_ports(&serial8250_reg, &serial8250_isa_devs->dev);
	
	ret = register_reboot_notifier(&vsc_serial_notifier);
	if(ret)
        {
                printk(KERN_ERR "vsc452_serial_notifier : cannot register reboot notifier (err=%d)\n", ret);
        }


#if 0
	ret = platform_driver_register(&serial8250_isa_driver);
	if (ret == 0)
		goto out;
	else
		printk(KERN_ERR "platform_driver_register failure in serial8250_init, aborting...\n");

	platform_device_unregister(serial8250_isa_devs);
 unreg:
	uart_unregister_driver(&serial8250_reg);
#endif	
 out:
	return ret;
}

static void __exit serial8250_exit(void)
{

	int i = 0;
#if 0
	struct platform_device *isa_dev = serial8250_isa_devs;

	/*
	 * This tells serial8250_unregister_port() not to re-register
	 * the ports (thereby making serial8250_isa_driver permanently
	 * in use.)
	 */
	 
	serial8250_isa_devs = NULL;

	platform_driver_unregister(&serial8250_isa_driver);
	platform_device_unregister(isa_dev);
#endif

	
	for (i = 0; i < UART_NR; i++) {
	//	struct uart_8250_port *up = &serial8250_ports[i];

	//	uart_remove_one_port((&serial8250_reg), &up->port);
		uart_remove_one_port((&serial8250_reg),&serial8250_ports[i].port );
	}

	uart_unregister_driver(&serial8250_reg);
	unregister_reboot_notifier(&vsc_serial_notifier);
}


/*********************************************************************************/

module_init(serial8250_init);
module_exit(serial8250_exit);

EXPORT_SYMBOL(serial8250_suspend_port);
EXPORT_SYMBOL(serial8250_resume_port);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Generic 8250/16x50 serial driver $Revision: 1.90 $");

//module_param(share_irqs, uint, 0644);
MODULE_PARM_DESC(share_irqs, "Share IRQs with other non-8250/16x50 devices"
	" (unsafe)");

MODULE_ALIAS_CHARDEV_MAJOR(TTY_MAJOR);
