/*************************************************************
* Module		: peci driver module
*       This driver implements PECI Rev 1.0. 
*
*************************************************************/
/* this enables dbg prints */
#define DEBUG_ENABLE  1

#if DEBUG_ENABLE == 1
#define dbgprint         printk
#else
#define dbgprint
#endif

/* Version Number of this module */
#define SAMPLE_MAJOR	1
#define SAMPLE_MINOR	0

#include <linux/module.h>
#include <linux/kernel.h>	/* printk() 		*/
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ide.h>
#include <linux/delay.h>

#include "peci.h"

#define PECI_MAJOR	45
#define PECI_DEV_NAME   "PECI"

/*Control reg bits for domain 0(0-3) and domain 1(4-7)*/
#define DOMAIN_0_BASE		0
#define DOMAIN_1_BASE		4
#define MAX_DOMAINS             2

#define PECI_STATUS_ERROR       0x30

/*PECI control ,status and data register*/
#define PECI_CTRL_REG           (*(volatile unsigned long*)0xB0201080)
#define PECI_STATUS_REG         (*(volatile unsigned long*)0xB0201084)
#define PECI_DATA_REG           (*(volatile unsigned long*)0xB0201088)

/*PECI Command execution time- roughly given based (there is nothing specified in the spec) */
#define PECI_INTERVAL   200

/*Write and read control register macros*/
#define WRITE_CTRL_REG(DATA)    PECI_CTRL_REG = (DATA)
#define READ_CTRL_REG(DATA)     (DATA) = PECI_CTRL_REG
#define CLEAR_CTRL_REG()        PECI_CTRL_REG = 0

/*Write and read status register macros*/
#define WRITE_STATUS_REG(DATA)    PECI_STATUS_REG = (DATA)
#define READ_STATUS_REG(DATA)     (DATA) = PECI_STATUS_REG
#define CLEAR_STATUS_REG()        PECI_STATUS_REG=0x0F

/*Write and read data register macros*/
#define READ_DATA_REG(DATA)     (DATA) = PECI_DATA_REG
#define CLEAR_DATA_REG()        PECI_DATA_REG = 0

/*PECI device errors return by Status registers (PECI Controller)*/
#define PECI_DEV_NO_ERROR       0x00
#define PECI_DEV_NEGO_FAILED    0x01
#define PECI_DEV_MISSING_NO_REP 0x02

#define PECI_RETRIES    5

/* Prototypes */
static int peci_open(struct inode * inode, struct file * filp);
static int peci_ioctl(struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg);
static int peci_init(void);
static int detect_cpu_presence ( unsigned char cpu_index, unsigned char *bpresence, int domain, unsigned int *ctrl_mask);
static int scan_temperature(unsigned char cpu_index, unsigned char domain, unsigned short *ptemp, unsigned char *pstatus);

static struct file_operations peci_fops = {
	owner:		THIS_MODULE,
	read:		NULL,
	write:		NULL,
	ioctl:		peci_ioctl,
	open:		peci_open,
	release:	NULL,
};

/* semaphore to provide mutex */
struct semaphore g_sem;

/*
 * waitForReply
 *
 * Waits for the command for the specific CPU to complete.
 *
 * input : mask - Bit pattern for the CPU that received the most recent command.
 *
 * returns 0 on success , 
 *         -1 if command times-out. 
 */
static int waitForReply(unsigned char mask) 
{
        int retval = -1;
        int complete = 0;
	volatile int ctrl_reg = 0;
        int retries = 0;

        // Loop until the command completes, or a timeout occurs.
        while (!complete) {
                // Delay before reading control register
                udelay(PECI_INTERVAL);
                READ_CTRL_REG (ctrl_reg);

                // If CPU bit is 0,then command completed.
                if ((ctrl_reg & mask) == 0) {
                        complete = 1;
                        retval = 0;
                }
                else {
                        // If CPU bit is set, command is not done.
                        // Increment a timeout counter and try again.
                        if (retries++ > PECI_RETRIES) {
                                complete = 1;
                        }
                }
        }

        return retval;
        
}

/*driver initialization function*/
static int
init_peci_module(void)
{
	int 	err;

	printk(KERN_ALERT "PECI  Driver Version %d.%d\n",SAMPLE_MAJOR,SAMPLE_MINOR);
	printk(KERN_ALERT "Copyright (c) 2008 American Megatrends Inc.\n");

	/* ----  Register the character device ------------------- */
	if ((err = register_chrdev(PECI_MAJOR, PECI_DEV_NAME, &peci_fops)) < 0)
	{
		printk (KERN_ERR "failed to register device (err %d)\n", err);
		return -ENODEV;
	}

	/*Clear all the peci registers*/	
	peci_init();

	/* Initialize the semaphore */
	init_MUTEX(&g_sem);
	
	return 0;
}

/*driver exit function*/
static void
cleanup_peci_module(void)
{
        int complete = 0;
        int counter = 0;
        
        printk(KERN_ALERT "Unloading PECI Module ..............\n");

        // If another process has the semaphore locked, delay before removing module
        while (!complete) {
                // Make sure semaphore is available before exiting.
                if (down_trylock(&g_sem)) {
                        // Call failed.  Wait for awhile before exiting.
                        udelay(PECI_INTERVAL);
                        if (counter++ > PECI_RETRIES) {
                                complete = 1;
                        }
                }
                else {
                        
                        complete = 1;
                }
        }

        
	/* uninitialize semaphore */

        /* unregister the device */
        unregister_chrdev(PECI_MAJOR, PECI_DEV_NAME);
	
        printk(KERN_ALERT "PECI module unloaded sucessfully.\n");
}
   

/* driver open function*/
static int peci_open(struct inode * inode, struct file * filp)
{
        // Note:  I'm not sure what the flag is checking.
        // I can't find the constant defined anywhere, but
        // I'm leaving it as-is since the code is replicated
        // in other drivers.
        if (filp != NULL) {
                if ((filp->f_flags & 010000000) == 0) {
                        filp->f_op =&peci_fops;
                }
        }
	return  0;
}

/*do the driver initializations here*/
static int peci_init( void )
{
	/* clear all registers*/	
	CLEAR_CTRL_REG();
	CLEAR_STATUS_REG();
	CLEAR_DATA_REG();
	return 0;
}


/*peci ioctl handler*/
static int peci_ioctl(struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg)
{
	ioctl_peci_data_t 	pecidata;
        unsigned int            mask;

	if ( copy_from_user(&pecidata, (void*)arg, sizeof(ioctl_peci_data_t)) )
	{
                return -EFAULT;
	}
                
	down(&g_sem);
	switch ( cmd )
	{
		case PECI_CPU_SCAN:
			pecidata.retvalue = detect_cpu_presence ( pecidata.cpu_index,&pecidata.status,pecidata.domain, &mask);	
		break;
		case PECI_GET_TEMP:
			pecidata.retvalue = scan_temperature ( pecidata.cpu_index, pecidata.domain, 
								&pecidata.temp, &pecidata.status );
		break;			
		default:
			dbgprint (" PECI:Invalid command\n");
			pecidata.retvalue = PECI_INVALID_CMD;
		break;
	}

	up(&g_sem);
	if ( copy_to_user((void*)arg,(void*)&pecidata, sizeof(ioctl_peci_data_t)) )
	{
		return -EFAULT;
	}
	return pecidata.retvalue;
}

/*
 * indexToBitPattern
 *
 * Converts a CPU/Domain pair to the proper bit pattern for command/status.
 *
 * input : cpu_index - Zero-based index for the CPU.
 *         domain    - Zero-based index for the domain.
 *
 * output : bitPattern - Bit-pattern for cpu/domain.
 *
 * returns PECI_NO_ERROR on success , 
 *         PECI_INVALID_PARAMS if any input parameter is invalid. 
 *
 */
int indexToBitPattern(unsigned char cpu_index, int domain, unsigned int *bitPattern)
{
	unsigned int domain_base = 0;
        int retVal = PECI_INVALID_PARAMS;

        // Validate CPU Index, domain, and bit_pattern
	if ((cpu_index < MAX_CPU_COUNT) && (domain < MAX_DOMAINS) && (domain >= 0) 
                && (bitPattern != NULL) ) {

                *bitPattern = 0;

                // Each domain is assigned 4-bits of the output.  Select proper nibble
                // based on domain.
                if (0 == domain) {
		        domain_base = DOMAIN_0_BASE;
	        } 
                else {
		        domain_base = DOMAIN_1_BASE;
	        }

                // Shift bit for CPU index in the proper nibble.
	        *bitPattern = ( 1 << (cpu_index + domain_base) );
                retVal = PECI_NO_ERROR;
	}

        return retVal;
}

/*
 * detect_cpu_presence
 *
 * Determines if the input cpu/domain is present.
 *
 * input : cpu_index - Zero-based index for the CPU.
 *         domain    - Zero-based index for the domain.
 *
 * output : bpresence - Status of cpu present check..
 *          ctrl_mask - bit-pattern for the cpu/domain pair.
 *
 * returns PECI_NO_ERROR on success , 
 *         PECI_INTERNAL_ERROR if check doesn't complete due to any error. 
 *
 */
static int detect_cpu_presence ( unsigned char cpu_index, unsigned char *bpresence, int domain, unsigned int *ctrl_mask)
{
	unsigned int ctrl_reg 	 = 0;
	unsigned int status_reg  = 0;
        unsigned int status_mask = 0;
        int retval = PECI_INTERNAL_ERROR;
        unsigned char byteReg = 0;

        // Verify that output parameters have been allocated.
        if ((bpresence != NULL) && (ctrl_mask != NULL))
        {
                *bpresence = 0;

                // Compute the bit-pattern for the cpu/domain pairing.
                retval = indexToBitPattern(cpu_index, domain, &ctrl_reg);
                if (retval == PECI_NO_ERROR) {

                        byteReg = ctrl_reg & 0xFF;

                        // Clear command/status registers.
	                CLEAR_STATUS_REG();
	                CLEAR_CTRL_REG();

                        // Write the command to PECI to see if CPU exists.
	                WRITE_CTRL_REG(byteReg);
                        *ctrl_mask = byteReg; 

                        // Wait for the command to complete.
                        if (waitForReply(byteReg) == 0) {

                                // Read status register
		                READ_STATUS_REG(status_reg);
                                status_mask = ( 1 << cpu_index);
                        
                                // First, see if the command had an error.
		                if ((status_reg & PECI_STATUS_ERROR) == 0) {
                                        // Next, see if the CPU is present.
                                        if ((status_reg & 0x0F) == status_mask) {
			                        *bpresence = 1;
                                        }
		                }
                                else {
                                        // Error reading status reg. Return an error.
                                        dbgprint(KERN_DEBUG "Detect: PECI Status Error: %x\n", status_reg & 0xFF);
                                        retval = PECI_INTERNAL_ERROR;
                                }
                        }
                        else {
                                // Command timed-out. Return an error
                                dbgprint (KERN_DEBUG "Detect: PECI: Busy  %d\n",byteReg);
		                retval = PECI_INTERNAL_ERROR;
	                }
                }
        }
        
	return retval;
}

/*
 * Scan Temperature
 * input : cpu_index, domain, temp, status
 * returns PECI_NO_ERROR on success, PECI_INTERNAL_ERROR, PECI_CPU_NOT_FOUND and PECI_INVALID_PARAMS on error.
 */
 
/*
 * scan_temperature
 *
 * Read the temperature for the givenj CPU/Domain pairing.
 *
 * input : cpu_index - Zero-based index for the CPU.
 *         domain    - Zero-based index for the domain.
 *
 * output : ptemp   - CPU/Domain temperature
 *          pstatus - Command status.
 *
 * returns PECI_NO_ERROR on success , 
 *         PECI_INTERNAL_ERROR if check doesn't complete due to any error. 
 *
 */
static int scan_temperature ( unsigned char cpu_index, unsigned char domain, unsigned short *ptemp, unsigned char *pstatus )
{
	unsigned int ctrl_mask =0;
	unsigned int ctrl_reg =0;
	unsigned int status_reg =0;
	unsigned char presence =0;
	int retvalue =0;

        // Verify that output parameters have been allocated.
        if ((ptemp != NULL) && (pstatus != NULL))
        {
                *pstatus = 0;

                // Determine if CPU is present.
	        retvalue =  detect_cpu_presence ( cpu_index, &presence, domain, &ctrl_mask); 
	        if ( PECI_NO_ERROR == retvalue )
	        {
	                if (presence) {
                                // CPU is present.
                                // Order a temperature scan.
                                CLEAR_CTRL_REG();        
                                WRITE_CTRL_REG(ctrl_mask);

                                // Wait for temperature scan to complete.
                                if (waitForReply(ctrl_mask & 0xFF) == 0) {

                                        // First, read status register to ensure that
                                        // no command error is reported.
                                        READ_STATUS_REG(status_reg);
                                        if ((status_reg & PECI_STATUS_ERROR) == 0) {
                                                // No errors were reported.  Read temperature.
                                                READ_DATA_REG(*ptemp);
                                        }
                                        else {
                                                // A command error was reported.
                                                // Return an error.
                                                retvalue = PECI_INTERNAL_ERROR;
                                                dbgprint(KERN_DEBUG "Scan: PECI Status Error: %x\n", status_reg & 0xFF);
                                        }
                                        *pstatus = status_reg;
                                }
                                else {
                                        // Temperature scan timed-out. Return an error.
                                        dbgprint ("Scan: PECI: Busy  %d\n",ctrl_reg);
                                        retvalue = PECI_INTERNAL_ERROR;
                                }
                        
	                }
                        else {
                                // CPU is not present. 
                                // Return an error.
                                retvalue = PECI_CPU_NOT_FOUND;
                        }
	        }
        }
                        
	return retvalue;
}

/*--------------------------- Module Information Follows --------------------------*/
module_init (init_peci_module);
module_exit (cleanup_peci_module);

MODULE_AUTHOR("American Megatrends Inc");
MODULE_DESCRIPTION("This is PECI driver module.");
MODULE_LICENSE("GPL");

