#include <config.h>
#include <common.h>
#include <net.h>
#include <asm/addrspace.h>
#include <vsc452/vsc452regs.h>
#include <vsc452/vsc452MAC.h>

#define REG_TDC    0x40
#define REG_RDC    0x48	 

inline 
void pioSt32(uint32_t addr, uint32_t val)
{
	*((volatile uint32_t*)addr) = val;
}

inline 
void pioLd32(uint32_t addr, uint32_t * valPtr)
{
	*valPtr = *((volatile uint32_t*)addr);
}

#define DBG_STR(argMask,argStr) printf(argStr);
/****************************************************************************
** MAC_Driver_Init : Initialize the MAC controller
****************************************************************************/

int 
MAC_Driver_Init ( uint32_t baseAddr, uint8_t * macAddress )
{
   // reset the MAC
	pioSt32 ( baseAddr + REG_MAC1, MAC1_SOR | MAC1_PEMCSRR | MAC1_PERFUNR | MAC1_PEMCSTR | MAC1_PETFUNR );
	pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS | TRCR_RFR | TRCR_RCDR | TRCR_TFR | TRCR_TCDR );

    // turn off reset
	pioSt32 ( baseAddr + REG_MAC1, 0 );
	pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS);

    // turn on huge frame enable, full-duplex, CRC enable
	pioSt32 ( baseAddr + REG_MAC2,  MAC2_FD | MAC2_CE | MAC2_ADPE | MAC2_PCE);

	if(macAddress != NULL)
	{
    	// set MAC address
	    pioSt32 ( baseAddr + REG_SA0, ( macAddress[0] << 8 ) + macAddress[1] );
    	pioSt32 ( baseAddr + REG_SA1, ( macAddress[2] << 8 ) + macAddress[3] );
	    pioSt32 ( baseAddr + REG_SA2, ( macAddress[4] << 8 ) + macAddress[5] );
	}

    // set to PHY mode, 100 Mb/s
    //pioSt32 ( baseAddr + REG_SUPP, SUPP_PM | SUPP_SPEED );
    pioSt32 ( baseAddr + REG_SUPP, SUPP_SPEED );

    // set Host Clock divided by 20
    pioSt32 ( baseAddr + REG_MCFG, 0x06 << 2 );

	// KMS we're polling, so no interrupts required. Also USB interrupts which 
	// we do use and enet interrupts are OR'd which my current ISR architecture doesn't support.
	// 
    // enable all interrupts except TFAE, RFAE, and EOF
    //pioSt32 ( baseAddr + REG_ISE, ~( ISE_TFAE | ISE_RFAE | ISE_EOF ) );
	// 

	// Clear all of the interrupt status bits
	pioSt32(baseAddr + REG_IS, 0xffffffff);
	pioSt32 ( baseAddr + REG_TDC, 0x00000000 );
    //pioSt32 ( KSEG1ADDR(SYS_MAC1_BASE) + REG_TDC, 0x00000000);

	pioSt32 ( baseAddr + REG_ISE, 0x00000000 );
    //pioSt32 (KSEG1ADDR(SYS_MAC1_BASE) + REG_ISE, 0x00000000 );

	pioSt32 ( baseAddr + REG_RDC, 0x00000000 );
    //pioSt32 ( KSEG1ADDR(SYS_MAC1_BASE) + REG_RDC, 0x00000000);
	
    // turn on receive enable
	pioSt32 ( baseAddr + REG_MAC1, MAC1_RE );

    return ( 0 );
}


/****************************************************************************
** MAC_Driver_Send : Send outgoing Ethernet Packet
****************************************************************************/

int MAC_Driver_Send (  uint32_t baseAddr, uint8_t * pktData, int pktLen )
{
	uint32_t* dataPtr = (uint32_t*)pktData;

    // verify pktLen
    if ( ( pktLen < 8 ) || ( pktLen > 2048 ) )
    {
        DBG_STR ( -1, "\r\nError - MAC_Driver_Send: Invalid Packet Length\r\n" );
        return ( 1 );
    }

    // pad out to the nearest 8 byte boundary
    if (  pktLen & 0x3 )
    {
        pktLen += 8;
        pktLen &= ~0x3;
    }

    // place first packet (dword) in the FIFO
    pioSt32 ( baseAddr + REG_TFC, TFC_SOP );
    pioSt32 ( baseAddr + REG_TFD, *dataPtr );
    pktLen -= 4;
    dataPtr++;

    // place middle packets (dwords) in the FIFO
    pioSt32 ( baseAddr + REG_TFC, TFC_MID );
    while ( pktLen > 4 )
    {
        pioSt32 ( baseAddr + REG_TFD, *dataPtr );
        pktLen -= 4;
        dataPtr++;
    }

    // place last packet (dword) in the FIFO
    switch ( pktLen )
    {
        case 1 :
                pioSt32 ( baseAddr + REG_TFC, TFC_EOP1 );
                break;
        case 2 :
                pioSt32 ( baseAddr + REG_TFC, TFC_EOP2 );
                break;
        case 3 :
                pioSt32 ( baseAddr + REG_TFC, TFC_EOP3 );
                break;
        case 4 :
                pioSt32 ( baseAddr + REG_TFC, TFC_EOP4 );
                break;
        default :
                DBG_STR ( -1, "MAC_Driver_Send: switch( pktLen ) error\r\n" );
                break;
    }
    pioSt32 ( baseAddr + REG_TFD, *dataPtr  );

    // Transmit the packets
    pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS | TRCR_TSFS );
    pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS );

    return ( 0 );
}


/****************************************************************************
** MAC_Driver_Rcv : Receive one Ethernet Packet
****************************************************************************/

int 
MAC_Driver_Rcv ( uint32_t baseAddr, uint8_t * pktData )
{
    uint32_t  * buf      = (uint32_t*)pktData;
    uint32_t    data;
    uint32_t    code;
    uint32_t    done;
    int       pktLen   = 0;

    /* Get Last Received Byte Count */
    pioLd32 ( baseAddr + REG_RSV, &data );
//	printf("Expected : 0x%lx \n",data & 0xFFFF);


    done   = 0;
    pktLen = 0;

    do
    {
   		// read in the data from FIFO
        pioLd32 ( baseAddr + REG_RFD, &data );
		if (pktLen <= 1536)
        	 *buf = data;

        // process the corresponding code
        pioLd32 ( baseAddr + REG_RFC, &code );
		code = code & 0xF;
        switch ( code )
        {
                case RFC_SOP :
                    buf += 1;
                    pktLen += 4;
                    break;
                case RFC_MID :
                    buf += 1;
                    pktLen += 4;
                    break;
                case RFC_EOP1 :
                    pktLen += 1;
                    done = 1;
                    break;
                case RFC_EOP2 :
                    pktLen += 2;
                    done = 1;
                    break;
                case RFC_EOP3 :
                    pktLen += 3;
                    done = 1;
                    break;
                case RFC_EOP4 :
                    pktLen += 4;
                    done = 1;
                    break;
                default :
              //      printf("MAC_Driver_Rcv: RFC  error, code=0x%x\n",code );
                    done = 1;
                    break;
            }
    } while ( !done );
//	printf("Got : 0x%lx \n",pktLen);
    return ( pktLen );
}


/****************************************************************************
** MAC_Driver_Svc_Int : Handle pending interrupts
****************************************************************************/



int eth_init(bd_t *bis)
{
	uint8_t macAddr[] = {0x00, 0x04, 0x03, 0x02, 0x01, 0x55};
	char * eth1addr;
	char *s;
	char *e;
	int i =0;
	unsigned int eth1base = 0;
	printf("Mac Addr = %02X:%02X:%02X:%02X:%02X:%02X\n",
				bis->bi_enetaddr[0],
				bis->bi_enetaddr[1],
				bis->bi_enetaddr[2],
				bis->bi_enetaddr[3],
				bis->bi_enetaddr[4],
				bis->bi_enetaddr[5]);


	MAC_Driver_Init ( KSEG1ADDR(SYS_MAC0_BASE), bis->bi_enetaddr );
    /* In the uboot we are using only eth0, eth1 mac address initialized in case if we have eth1addr variable set*/
#if 1	
	//Set the MAC1 ethaddr
	eth1addr = getenv("eth1addr");
	if ( NULL == eth1addr )
		return 0;
	if (strncmp(eth1addr,"00:00:00:00:00:00",17) == 0)
	{
		return 0;
	}
	
	s= &eth1addr[0];
	for (i = 0; i < 6; ++i)
	{
		macAddr[i] = s ? simple_strtoul(s, &e, 16) : 0;
        if (s) s = (*e) ? e+1 : e;
    }

	eth1base = KSEG1ADDR(SYS_MAC1_BASE);
    pioSt32 ( eth1base + REG_SA0, ( macAddr[0] << 8 ) + macAddr[1] );
    pioSt32 ( eth1base + REG_SA1, ( macAddr[2] << 8 ) + macAddr[3] );
    pioSt32 ( eth1base + REG_SA2, ( macAddr[4] << 8 ) + macAddr[5] );
	MAC_Driver_Init ( eth1base, NULL );
#endif
	return 0;
}

int eth_send(volatile void *packet, int length)
{
	return MAC_Driver_Send (  KSEG1ADDR(SYS_MAC0_BASE), (uint8_t *)packet, length );
}

uint8_t rxBuff[2048];

int eth_rx(void)
{
	int pktLen = 0;
	uint32_t data;
	uint32_t baseAddr = KSEG1ADDR(SYS_MAC0_BASE);

	while (1)
	{
		pioLd32( baseAddr + REG_IS, &data );
		pioSt32(baseAddr + REG_IS, data);	   
		if (data & IS_RFO)
		{
    		pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS | TRCR_RFR|TRCR_RCDR);
    		pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS);
			printf("O");
//			printf("Rcv OverFlow\n");
			continue;
		}
		if (data & IS_RFE)
			break;
		pktLen = MAC_Driver_Rcv (baseAddr, rxBuff);			
		if(pktLen <=4)
		{
//			printf("EOF len = %d\n",pktLen);
			break;
		}
		if (pktLen > 1536)
		{
//			printf("TOO BIG pktLen = %d\n",pktLen);
			continue;
		}
		if (pktLen < 0x40)
		{
//			printf("TOO SMALL PktLen = %d\n",pktLen);
			continue;
		}
	//	printf("pktLen = %d\n",pktLen);
		NetReceive(rxBuff, pktLen);
	}
	
	return pktLen;
}

void eth_halt(void)
{
	unsigned long baseAddr = KSEG1ADDR(SYS_MAC0_BASE);
    pioSt32 ( baseAddr + REG_MAC1, 0 );
    pioSt32 ( baseAddr + REG_TRCR, TRCR_RMIIROS );
	return;
}

char *eth_get_name (void)
{
	return "452MAC";
}











