/*
 * (C) Copyright 2003
 * Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>

#define MAX_INTR 8

typedef unsigned long UINT32;

struct irq_action
{
	interrupt_handler_t *handler;
	void *arg;
	int count;
	int spurious;
};

static struct irq_action vectors[MAX_INTR];

void
SpuriousHandler(void *IntNum)
{
	printf("Spurious Intr %d received \n",(UINT32)IntNum);
}

void
ClearIntr(UINT32 IntNum)
{
	/* TODO */
	return;
}

void enable_interrupts(void)
{
	__asm__ __volatile__(".set mips32r2\nei");
}

int disable_interrupts(void)
{
	unsigned long prev;
	__asm__ __volatile__(".set mips32r2\ndi %0" : "=r" (prev));
	return prev & 0x1;
}


void
MaskIntr(int i)
{
	/* Clear Intrerrupt Bit in CP0 Status Register */
	UINT32 IntMask;
	if (i >=MAX_INTR)
		return;
	IntMask = (1 << i) << 10;
	IntMask = ~IntMask;

	__asm__ __volatile__(	\
	" .set push 	\n"		\
	" .set noreorder	\n"		\
	" .set mips32r2	\n"		\
	" mfc0 $8, $12	\n"		\
	" ehb			\n"		\
	" and  $8, %0	\n"		\
	" mtc0 $8, $12	\n"		\
	" .set pop		\n"		\
	:						\
	: "r" (IntMask)			\
	: "$8");

	return;
}

void
UnMaskIntr(int i)
{
	/* Set Intrerrupt Bit in CP0 Status Register */
	UINT32 IntMask;
	if (i >=MAX_INTR)
		return;
	IntMask = (1 << i) << 10;

	__asm__ __volatile__(	\
	" .set push 	\n"		\
	" .set noreorder\n"		\
	" .set mips32r2	\n"		\
	" mfc0 $8, $12	\n"		\
	" ehb			\n"		\
	" or   $8, %0	\n"		\
	" mtc0 $8, $12	\n"		\
	" .set pop		\n"		\
	:						\
	: "r" (IntMask)			\
	: "$8");

	return;
}

int
interrupt_init(void)
{
	UINT32 i;
	/* Initialize to Dummy Handler */
	for (i=0;i<MAX_INTR;i++)
	{
		vectors[i].handler = NULL;
		vectors[i].arg     = (void*)i;
		vectors[i].count   = 0;
		vectors[i].spurious = 0;
	}

	/* Clear Pending Intr */  
	for (i=0;i<MAX_INTR;i++)
		ClearIntr(i);

	/* Clear Timer Intr */
	/* Writing to Compare register CP0 11, clear the interrupt */
	i=0;
	asm volatile ("mtc0 %0, $11" : : "r" (i));

	/* Disable All Intr*/
	for (i=0;i<MAX_INTR;i++)
		MaskIntr(i);

	enable_interrupts();
	return 0;
}

void 
HandleIntr(unsigned long IntrStatus)
{
	unsigned long IntNum=0;

	printf("Intr got\n");
	while (IntrStatus)
	{
		if (IntrStatus & 0x1)
		{
			if (vectors[IntNum].handler == NULL)
			{
				vectors[IntNum].spurious++;
				SpuriousHandler((void *)IntNum);
			}
			else
			{
				vectors[IntNum].count++;
				printf("Intr got for %d\n",IntNum);
				printf("Calling 0x%lx\n",vectors[IntNum].handler);
				(*vectors[IntNum].handler)(vectors[IntNum].arg);
			}
			ClearIntr(IntNum);
		}
		IntNum++;		
		IntrStatus >>=1;

		/* Just a precaution check */
		if (IntNum >= MAX_INTR)
			break;
	}
	return;
}

void
irq_install_handler(int IntNum, interrupt_handler_t *handler, void *arg)
{
	printf("called for %d 0x%lx\n",IntNum,handler);
	if (IntNum >=MAX_INTR)
	{
		printf("ERROR: Unsupported INTR Number %d\n",IntNum);
		return;
	}

	if (handler == NULL)
	{
		printf("WARNING: NULL Handler. Freeing Interrupt %d\n",IntNum);
		irq_free_handler(IntNum);
		return;
	}

	if (vectors[IntNum].handler != NULL)
	{
		if (vectors[IntNum].handler == handler)
		{
			printf("WARNING: Same vector is installed for INT %d\n",IntNum);
			return;
		}
		printf("Replacing vector for INT %d\n",IntNum);
	}

	disable_interrupts();
	vectors[IntNum].handler = handler;
	vectors[IntNum].arg = arg;
	enable_interrupts();

	/* Enable the specfic Intr*/
	UnMaskIntr(IntNum);

	return;
}


void
irq_free_handler(int IntNum)
{
	/* Disable  the specfic Intr */
	MaskIntr(IntNum);

	disable_interrupts();
	vectors[IntNum].handler = NULL;
	vectors[IntNum].arg = (void *)IntNum;
	enable_interrupts();
	return;

}


#if (CONFIG_COMMANDS & CFG_CMD_IRQ)

int
do_irqinfo(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int vec;

	printf ("\nInterrupt-Information:\n");
	printf ("Num  Routine   Arg       Count  Spurious\n");

	for (vec=0; vec<MAX_INTR; vec++)
	{
		printf ("%02d   %08lx  %08lx      %d         %d\n",
				vec,
				(ulong)vectors[vec].handler,
				(ulong)vectors[vec].arg,
				vectors[vec].count,
				vectors[vec].spurious);
	}
	return 0;
}

#endif  /* CONFIG_COMMANDS & CFG_CMD_IRQ */
